#include <process.h>
#include <assert.h>
#include <windows.h>
#include <wingdi.h>
#include <math.h>

#include "dynamic_funcs.h"
#include "ac_string_util.h"
#include "fontgen.h"
#include "fontchar.h"
#include "unicode.h"
#include "acimg.h"

#define CLR_BORDER 0x007F00ul
#define CLR_UNUSED 0xFF0000ul

CFontGen::CFontGen()
{
	fontChanged       = true;
	isWorking         = false;
	stopWorking       = false;
	status            = 0;
	arePagesGenerated = false;

	fontName        = "Arial";
	charSet         = ANSI_CHARSET;
	fontSize        = 32;
	aa              = 1;
	useSmoothing    = true;
	isBold          = false;
	isItalic        = false;
	useUnicode      = true;
	paddingLeft     = 0;
	paddingRight    = 0;
	paddingUp       = 0;
	paddingDown     = 0;
	spacingHoriz    = 1;
	spacingVert     = 1;
	disableBoxChars = true;
	scaleH          = 100;

	outWidth           = 256;
	outHeight          = 256;
	outBitDepth        = 8;
	fourChnlPacked     = false;
	textureFormat      = "tga";
	textureCompression = 0;
	fontDescFormat     = 0;

	outlineThickness   = 0;
	is1ChnlEncoded     = false;

	memset(selected, 0, sizeof(selected));
	memset(chars, 0, sizeof(chars));
}

CFontGen::~CFontGen()
{
	assert(!isWorking);

	ClearPages();

	ClearIconImages();
}

void CFontGen::ClearIconImages()
{
	assert(!isWorking);
	arePagesGenerated = false;

	for( unsigned int n = 0; n < iconImages.size(); n++ )
	{
		if( iconImages[n] )
			delete iconImages[n];
	}

	iconImages.resize(0);
}

int CFontGen::AddIconImage(const char *file, int id)
{
	assert(!isWorking);
	arePagesGenerated = false;

	// Load the image file
	acImage::Image rawImg;
	int r = LoadImageFile(file, rawImg);
	if( r < 0 )
		return -r;

	acImage::Image rgbImg;
	acImage::ConvertToARGB(rgbImg, rawImg);

	cImage *image = new cImage(rgbImg.width, rgbImg.height);
	memcpy(image->pixels, rgbImg.data, rgbImg.width*rgbImg.height*4);

	SIconImage *i = new SIconImage;
	i->fileName = file;
	i->id       = id;
	i->image    = image;

	iconImages.push_back(i);

	return 0;
}

int CFontGen::DeleteIconImage(int id)
{
	assert(!isWorking);
	arePagesGenerated = false;

	// Find the icon image for removal
	bool isFound = false;
	for( int n = 0; n < (signed)iconImages.size(); n++ )
	{
		if( iconImages[n]->id == id )
		{
			isFound = true;

			SIconImage *img = iconImages[n];
			if( n < (signed)iconImages.size() - 1 )
				iconImages[n] = iconImages[iconImages.size()-1];
			iconImages.pop_back();

			delete img;

			break;
		}
	}

	return isFound ? 0 : -1;
}

int CFontGen::UpdateIconImage(int oldId, int id, const char *file)
{
	assert(!isWorking);
	arePagesGenerated = false;

	// Find the icon image for update
	bool isFound = false;
	for( int n = 0; n < (signed)iconImages.size(); n++ )
	{
		if( iconImages[n]->id == oldId )
		{
			isFound = true;
			iconImages[n]->id = id;

			if( iconImages[n]->fileName != file )
			{
				// Load the image file
				acImage::Image rawImg;
				int r = LoadImageFile(file, rawImg);
				if( r < 0 )
					return -r;

				acImage::Image rgbImg;
				acImage::ConvertToARGB(rgbImg, rawImg);

				cImage *image = new cImage(rgbImg.width, rgbImg.height);
				memcpy(image->pixels, rgbImg.data, rgbImg.width*rgbImg.height*4);

				iconImages[n]->fileName = file;
				delete iconImages[n]->image;
				iconImages[n]->image = image;
			}

			break;
		}
	}

	return isFound ? 0 : -1;
}

int CFontGen::GetIconImageCount()
{
	return iconImages.size();
}

int CFontGen::GetIconImageInfo(int n, string &filename, int &id)
{
	if( n < 0 || n >= (signed)iconImages.size() )
		return -1;

	filename = iconImages[n]->fileName;
	id = iconImages[n]->id;

	return 0;
}

void CFontGen::Abort()
{
	stopWorking = true;

	while( isWorking )
		Sleep(10);
}

bool CFontGen::IsUsingUnicode()
{
	return useUnicode;
}

int CFontGen::SetUseUnicode(bool set)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	if( useUnicode != set )
		fontChanged = true;
	useUnicode = set;

	return 0;
}

bool CFontGen::Is4ChnlPacked()
{
	return fourChnlPacked;
}

int CFontGen::Set4ChnlPacked(bool set)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	fourChnlPacked = set;

	return 0;
}

string CFontGen::GetTextureFormat()
{
	return textureFormat;
}

int CFontGen::SetTextureFormat(string &format)
{
	textureFormat = format;

	return 0;
}

int CFontGen::GetTextureCompression()
{
	return textureCompression;
}

int CFontGen::SetTextureCompression(int compression)
{
	textureCompression = compression;

	return 0;
}

int CFontGen::IsSubsetSelected(int subset)
{
	bool allChecked = true;
	bool someChecked = false;

	for( int n = UnicodeSubsets[subset].range[0]; n <= UnicodeSubsets[subset].range[1]; n++ )
	{
		if( !disabled[n] )
		{
			if( selected[n] )
				someChecked = true;
			else
				allChecked = false;
		}
	}

	if( allChecked && someChecked )
		return 2;
	else if( someChecked )
		return 1;

	return 0;
}

int CFontGen::SelectSubset(int subset, bool select, bool entireRange)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	// Set the new selection state
	if( !entireRange )
	{
		for( int n = UnicodeSubsets[subset].range[0]; n <= UnicodeSubsets[subset].range[1]; n++ )
			SetSelected(n, select);
	}
	else
	{
		int set = UnicodeSubsets[subset].set;
		for( int s = 0; s < numUnicodeSubsets; s++ )
		{
			if( UnicodeSubsets[s].set == set )
			{
				for( int n = UnicodeSubsets[s].range[0]; n <= UnicodeSubsets[s].range[1]; n++ )
					SetSelected(n, select);
			}
		}
	}

	return 0;
}

int CFontGen::GetStatus()
{
	return status;
}

int CFontGen::GetStatusCounter()
{
	return counter;
}

int CFontGen::GetNumCharsSelected()
{
	return numCharsSelected;
}

int CFontGen::GetNumCharsAvailable()
{
	return numCharsAvailable;
}

bool CFontGen::DidNotFit(int charIdx)
{
	return noFit[charIdx];
}

bool CFontGen::IsDisabled(int charIdx)
{
	return disabled[charIdx];
}

bool CFontGen::IsSelected(int charIdx)
{
	return selected[charIdx];
}

HFONT CFontGen::CreateFont(int FontSize)
{
	if( FontSize == 0 ) FontSize = fontSize*aa;
	DWORD quality = useSmoothing ? ANTIALIASED_QUALITY : NONANTIALIASED_QUALITY;

	HFONT font = ::CreateFontA(FontSize, 0, 0, 0, isBold ? FW_BOLD : FW_NORMAL, isItalic ? TRUE : FALSE, 0, 0, charSet, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, quality, DEFAULT_PITCH, fontName.c_str());

	return font;
}


int CFontGen::SetSelected(int idx, bool set)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	if( disabled[idx] )
		return -1;

	if( selected[idx] != set )
	{
		selected[idx] = set;
		numCharsSelected += set ? 1 : -1;
	}

	return 0;
}

int CFontGen::Prepare()
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	// Update some properties that are exclusive
	if( fourChnlPacked && outBitDepth != 32 )
		Set4ChnlPacked(false);

	if( is1ChnlEncoded && outlineThickness == 0 )
		Set1ChnlEncoded(false);

	if( outlineThickness > 0 && fourChnlPacked && !is1ChnlEncoded )
		Set4ChnlPacked(false);

	ResetFont();

	return 0;
}

int CFontGen::SetBoxCharsDisabled(bool set)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	if( disableBoxChars != set )
		fontChanged = true;

	disableBoxChars = set;
	return 0;
}

int CFontGen::SetFontDescFormat(int format)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	fontDescFormat = format;
	return 0;
}

int CFontGen::SetOutBitDepth(int bitDepth)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	if( bitDepth != 8 && bitDepth != 32 )
		bitDepth = 8;

	outBitDepth = bitDepth;
	return 0;
}

int CFontGen::SetOutHeight(int height)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	outHeight = height;
	return 0;
}

int CFontGen::SetOutWidth(int width)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	outWidth = width;
	return 0;
}

int CFontGen::GetOutlineThickness()
{
	return outlineThickness;
}

int CFontGen::SetOutlineThickness(int thickness)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	outlineThickness = thickness;
	return 0;
}

bool CFontGen::Is1ChnlEncoded()
{
	return is1ChnlEncoded;
}

int CFontGen::Set1ChnlEncoded(bool set)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	is1ChnlEncoded = set;
	return 0;
}

int CFontGen::SetScaleHeight(int scale)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	scaleH = scale;
	return 0;
}

int CFontGen::SetSpacingHoriz(int space)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	spacingHoriz = space;
	return 0;
}

int CFontGen::SetSpacingVert(int space)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	spacingVert = space;
	return 0;
}

int CFontGen::SetPaddingUp(int pad)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	paddingUp = pad;
	return 0;
}

int CFontGen::SetPaddingDown(int pad)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	paddingDown = pad;
	return 0;
}

int CFontGen::SetPaddingLeft(int pad)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	paddingLeft = pad;
	return 0;
}

int CFontGen::SetPaddingRight(int pad)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	paddingRight = pad;
	return 0;
}

int CFontGen::SetUseSmoothing(bool set)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	useSmoothing = set;
	return 0;
}

int CFontGen::SetAntiAliasingLevel(int level)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	aa = level;
	return 0;
}

int CFontGen::SetItalic(bool set)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	isItalic = set;
	return 0;
}

int CFontGen::SetBold(bool set)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	isBold = set;
	return 0;
}

int CFontGen::SetFontSize(int fontSize)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	this->fontSize = fontSize;
	return 0;
}

int CFontGen::SetCharSet(int charSet)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	if( this->charSet != charSet )
		fontChanged = true;

	this->charSet = charSet;
	return 0;
}

int CFontGen::SetFontName(string &name)
{
	if( isWorking ) return -1;
	arePagesGenerated = false;

	if( fontName != name )
		fontChanged = true;

	fontName = name;
	return 0;
}

bool CFontGen::IsBoxCharsDisabled()
{
	return disableBoxChars;
}

int CFontGen::GetFontDescFormat()
{
	return fontDescFormat;
}

int CFontGen::GetOutBitDepth()
{
	return outBitDepth;
}

int CFontGen::GetOutHeight()
{
	return outHeight;
}

int CFontGen::GetOutWidth()
{
	return outWidth;
}

int CFontGen::GetScaleHeight()
{
	return scaleH;
}

int CFontGen::GetSpacingHoriz()
{
	return spacingHoriz;
}

int CFontGen::GetSpacingVert()
{
	return spacingVert;
}

int CFontGen::GetPaddingUp()
{
	return paddingUp;
}

int CFontGen::GetPaddingDown()
{
	return paddingDown;
}

int CFontGen::GetPaddingLeft()
{
	return paddingLeft;
}

int CFontGen::GetPaddingRight()
{
	return paddingRight;
}

bool CFontGen::IsUsingSmoothing()
{
	return useSmoothing;
}

bool CFontGen::IsItalic()
{
	return isItalic;
}

bool CFontGen::IsBold()
{
	return isBold;
}

int CFontGen::GetAntiAliasingLevel()
{
	return aa;
}

int CFontGen::GetFontSize()
{
	return fontSize;
}

int CFontGen::GetCharSet()
{
	return charSet;
}

cImage *CFontGen::GetPageImage(int page, int channel)
{
	pages[page]->GeneratePreviewTexture(channel);
	return pages[page]->GetPageImage();
}

int CFontGen::GetNumPages()
{
	return pages.size();
}

string CFontGen::GetFontName()
{
	return fontName;
}

// Internal
void CFontGen::ResetFont()
{
	if( fontChanged )
	{
		memset(disabled, 0, sizeof(disabled));
		memset(noFit, 0, sizeof(noFit));

		DetermineExistingChars();
	}
	fontChanged = false;
}

// Internal
void CFontGen::DetermineExistingChars()
{
	if( disableBoxChars && fGetGlyphIndicesA )
	{
		HDC dc = GetDC(0);

		HFONT font = CreateFont(10);
		HFONT oldFont = (HFONT)SelectObject(dc, font);

		numCharsAvailable = 0;
		numCharsSelected = 0;

		if( useUnicode )
		{
			int n;
			for( n = 0; n < (signed)numUnicodeSubsets; n++ )
				UnicodeSubsets[n].isAvailable = false;

			for( n = 0; n < 65536; n++ )
				disabled[n] = false;

			int subset = 0;
			for( n = 0; n < 65536; n++ )
			{
				unsigned short buf[2];
				buf[0] = n;
				buf[1] = 0;

				WORD idx;
				int r = fGetGlyphIndicesW(dc, buf, 1, &idx, GGI_MARK_NONEXISTING_GLYPHS);

				if( r == GDI_ERROR || idx == 0xFFFF )
					disabled[n] = true;
				else
				{
					// Determine which subset the glyph is in
					while( UnicodeSubsets[subset].range[1] < n )
						subset++;
					UnicodeSubsets[subset].isAvailable = true;

					numCharsAvailable++;
					if( selected[n] ) 
						numCharsSelected++;
				}
			}
		}
		else
		{
			int n;
			for( n = 0; n < 256; n++ )
				disabled[n] = false;

			for( n = 0; n < 256; n++ )
			{
				char buf[2];
				buf[0] = n;
				buf[1] = '\0';

				WORD idx;
				int r = fGetGlyphIndicesA(dc, buf, 1, &idx, GGI_MARK_NONEXISTING_GLYPHS);

				if( r == GDI_ERROR || idx == 0xFFFF )
					disabled[n] = true;
				else
				{
					numCharsAvailable++;
					if( selected[n] ) 
						numCharsSelected++;
				}
			}
		}

		SelectObject(dc, oldFont);
		DeleteObject(font);

		ReleaseDC(0, dc);
	}
}

// Internal
void CFontGen::ClearPages()
{
	for( int n = 0; n < (signed)pages.size(); n++ )
	{
		if( pages[n] ) delete pages[n];
		pages[n] = 0;
	}

	pages.clear();

	for( int n = 0; n < 65536; n++ )
	{
		if( chars[n] ) delete chars[n];
		chars[n] = 0;
	}
}

// Internal
int CFontGen::CreatePage()
{
	CFontPage *page = new CFontPage(this, pages.size(), outWidth, outHeight, spacingHoriz, spacingVert);
	page->SetPadding(paddingLeft, paddingUp, paddingRight, paddingDown);
	page->SetIntendedFormat(outBitDepth, fourChnlPacked, is1ChnlEncoded);
	pages.push_back(page);

	return pages.size() - 1;
}

// Internal
void CFontGen::InternalGeneratePages()
{
	if( arePagesGenerated )
	{
		status    = 0;
		isWorking = false;
		return;
	}

	ClearPages();

	bool didNotFit = false;
	memset(noFit, 0, sizeof(noFit));

	if( stopWorking )
	{
		status    = 0;
		isWorking = false;
		return;
	}

	const int maxChars = useUnicode ? 65536 : 256;

	// Add the imported images to the character list
	for( int n = 0; n < (signed)iconImages.size(); n++ )
	{
		int ch = iconImages[n]->id;
		chars[ch] = new CFontChar();
		chars[ch]->CreateFromImage(n, iconImages[n]->image);

		if( chars[ch]->height > 0 && chars[ch]->width > 0 )
		{
			if( (chars[ch]->height + paddingUp + paddingDown) > outHeight-spacingVert || 
				(chars[ch]->width + paddingRight + paddingLeft) > outWidth-spacingHoriz )
			{
				didNotFit = true;
				noFit[ch] = true;	

				// Delete the character again so that it isn't considered again
				delete chars[ch];
				chars[ch] = 0;
			}
		}

		if( stopWorking )
		{
			status    = 0;
			isWorking = false;
			return;
		}
	}

	// Draw each of the chars into individual images
	HFONT font = CreateFont(0);
	for( int n = 0; n < maxChars; n++ )
	{
		if( !disabled[n] && selected[n] )
		{
			// Unless the image is taken by an imported icon
			// Draw the character in a separate image
			// Determine the dimensions of the character
			if( chars[n] == 0 )
			{
				chars[n] = new CFontChar();
				chars[n]->DrawChar(font, n, scaleH, aa, useUnicode);
				if( outlineThickness )
					chars[n]->AddOutline(outlineThickness);

				if( chars[n]->height > 0 && chars[n]->width > 0 )
				{
					if( (chars[n]->height + paddingUp + paddingDown) > outHeight-spacingVert || 
						(chars[n]->width + paddingRight + paddingLeft) > outWidth-spacingHoriz )
					{
						didNotFit = true;
						noFit[n] = true;	

						// Delete the character again so that it isn't considered again
						delete chars[n];
						chars[n] = 0;
					}
				}
			}
			counter++;

			if( stopWorking )
			{
				status    = 0;
				isWorking = false;
				DeleteObject(font);
				return;
			}
		}
	}
	DeleteObject(font);

	// Build  a list of used characters
	status = 2;
	counter = 0;

	static CFontChar *ch[65536];
	int numChars = 0;
	for( int n = 0; n < maxChars; n++ )
	{
		if( chars[n] )
			ch[numChars++] = chars[n];
	}

	// Create pages until there are no more chars
	while( numChars > 0 )
	{
		int page = CreatePage();
		pages[page]->AddChars(ch, numChars);

		// Compact list
		for( int n = 0; n < numChars; n++ )
		{
			if( ch[n] == 0 )
			{
				// Find the last char
				for( numChars--; numChars > n; numChars-- )
				{
					if( ch[numChars] )
					{
						ch[n] = ch[numChars];
						ch[numChars] = 0;
						break;
					}
				}
			}
		}

		if( stopWorking )
		{
			status    = 0;
			isWorking = false;
			return;
		}
	}

	status    = 0;
	isWorking = false;
	arePagesGenerated = true;
}

// Internal
void CFontGen::GenerateThread(CFontGen *fontGen)
{
	fontGen->InternalGeneratePages();
}

int CFontGen::GeneratePages()
{
	if( isWorking ) return -1;

	// Set status refresh timer
	status = 1;
	counter = 0;

	isWorking         = true;
	stopWorking       = false;

	memset(noFit, 0, sizeof(noFit));

	_beginthread((void (*)(void*))GenerateThread, 0, this);	

	return 0;
}

int CFontGen::SaveFont(const char *szFile)
{
	if( isWorking ) return -1;

	// The pages must be generated first
	if( !arePagesGenerated ) return -1;

	// Create a memory dc
	HDC dc = CreateCompatibleDC(0);

	HFONT font = CreateFont(0);
	HFONT oldFont = (HFONT)SelectObject(dc, font);

	// Determine the size needed for the char
	int height, base;

	TEXTMETRIC tm;
	GetTextMetrics(dc, &tm);

	height = tm.tmHeight/aa;
	base = tm.tmAscent/aa;

	// Save the character attributes
	FILE *f;

	string filename = szFile;
	if( stricmp(filename.substr(filename.length() - 4).c_str(), ".fnt") == 0 )
		filename = filename.substr(0, filename.length() - 4);

	f = fopen((filename + ".fnt").c_str(), "wb");
	if( f == 0 )
		return -1;

	// Get the filename without path
	int r = filename.rfind('\\');
	string filenameonly;
	if( r != -1 )
		filenameonly = filename.substr(r+1);
		
	if( outBitDepth != 32 ) fourChnlPacked = false;
	int numPages = pages.size();

	if( fontDescFormat == 1 ) 
	{
		fprintf(f, "<?xml version=\"1.0\"?>\r\n");
		fprintf(f, "<font>\r\n");
		fprintf(f, "  <info face=\"%s\" size=\"%d\" bold=\"%d\" italic=\"%d\" charset=\"%s\" unicode=\"%d\" stretchH=\"%d\" smooth=\"%d\" aa=\"%d\" padding=\"%d,%d,%d,%d\" spacing=\"%d,%d\" outline=\"%d\"/>\r\n", fontName.c_str(), fontSize, isBold, isItalic, useUnicode ? "" : GetCharSetName(charSet).c_str(), useUnicode, scaleH, useSmoothing, aa, paddingUp, paddingRight, paddingDown, paddingLeft, spacingHoriz, spacingVert, outlineThickness);
		fprintf(f, "  <common lineHeight=\"%d\" base=\"%d\" scaleW=\"%d\" scaleH=\"%d\" pages=\"%d\" packed=\"%d\" encoded=\"%d\"/>\r\n", int(ceilf(height*float(scaleH)/100.0f)), int(ceilf(base*float(scaleH)/100.0f)), outWidth, outHeight, numPages, fourChnlPacked, is1ChnlEncoded);

		fprintf(f, "  <pages>\r\n");
		for( int n = 0; n < numPages; n++ )
			fprintf(f, "    <page id=\"%d\" file=\"%s_%02d.%s\" />\r\n", n, filenameonly.c_str(), n, textureFormat.c_str());
		fprintf(f, "  </pages>\r\n");
	}
	else if( fontDescFormat == 0 )
	{
		fprintf(f, "info face=\"%s\" size=%d bold=%d italic=%d charset=\"%s\" unicode=%d stretchH=%d smooth=%d aa=%d padding=%d,%d,%d,%d spacing=%d,%d outline=%d\r\n", fontName.c_str(), fontSize, isBold, isItalic, useUnicode ? "" : GetCharSetName(charSet).c_str(), useUnicode, scaleH, useSmoothing, aa, paddingUp, paddingRight, paddingDown, paddingLeft, spacingHoriz, spacingVert, outlineThickness);
		fprintf(f, "common lineHeight=%d base=%d scaleW=%d scaleH=%d pages=%d packed=%d encoded=%d\r\n", int(ceilf(height*float(scaleH)/100.0f)), int(ceilf(base*float(scaleH)/100.0f)), outWidth, outHeight, numPages, fourChnlPacked, is1ChnlEncoded);

		for( int n = 0; n < numPages; n++ )
			fprintf(f, "page id=%d file=\"%s_%02d.%s\"\r\n", n, filenameonly.c_str(), n, textureFormat.c_str());
	}
	else
	{
		// Write the magic word and file version
		fwrite("BMF", 3, 1, f);
		fputc(2, f); 

		// Write the info block
#pragma pack(push)
#pragma pack(1)
		struct infoBlock
		{
			int            blockSize;
			unsigned short fontSize;
			char           reserved:4;
			char           bold    :1;
			char           italic  :1;
			char           unicode :1;
			char           smooth  :1;
			unsigned char  charSet;
			unsigned short stretchH;
			char           aa;
			unsigned char  paddingUp;
			unsigned char  paddingRight;
			unsigned char  paddingDown;
			unsigned char  paddingLeft;
			unsigned char  spacingHoriz;
			unsigned char  spacingVert;
			unsigned char  outline;
			char           fontName[1];
		} info;
#pragma pack(pop)

		info.blockSize    = sizeof(info) + fontName.length();
		info.fontSize     = fontSize;
		info.reserved     = 0;
		info.bold         = isBold;
		info.italic       = isItalic;
		info.unicode      = useUnicode;
		info.smooth       = useSmoothing;
		info.charSet      = charSet;
		info.stretchH     = scaleH;
		info.aa           = aa;
		info.paddingUp    = paddingUp;
		info.paddingRight = paddingRight;
		info.paddingDown  = paddingDown;
		info.paddingLeft  = paddingLeft;
		info.spacingHoriz = spacingHoriz;
		info.spacingVert  = spacingVert;
		info.outline      = outlineThickness;

		fputc(1, f);
		fwrite(&info, sizeof(info)-1, 1, f);
		fwrite(fontName.c_str(), fontName.length()+1, 1, f);

		// Write the common block
#pragma pack(push)
#pragma pack(1)
		struct commonBlock
		{
			int blockSize;
			unsigned short lineHeight;
			unsigned short base;
			unsigned short scaleW;
			unsigned short scaleH;
			unsigned short pages;
			unsigned char  packed:1;
			unsigned char  encoded:1;
			unsigned char  reserved:6;
		} common; 
#pragma pack(pop)

		common.blockSize  = sizeof(common);
		common.lineHeight = int(ceilf(height*float(scaleH)/100.0f));
		common.base       = int(ceilf(base*float(scaleH)/100.0f));
		common.scaleW     = outWidth;
		common.scaleH     = outHeight;
		common.pages      = numPages;
		common.reserved   = 0;
		common.packed     = fourChnlPacked;
		common.encoded    = is1ChnlEncoded;

		fputc(2, f);
		fwrite(&common, sizeof(common), 1, f);

		// Write the page block
		fputc(3, f);
		int size = 4 + (filenameonly.length() + 4 + textureFormat.length() + 1)*numPages;
		fwrite(&size, sizeof(size), 1, f);

		for( int n = 0; n < numPages; n++ )
		{
			fprintf(f, "%s_%02d.%s", filenameonly.c_str(), n, textureFormat.c_str());
			fputc(0, f);
		}
	}

	const int maxChars = useUnicode ? 65536 : 256;

	// Count the number of characters that will be written
	int numChars = 0;
	int n;
	for( n = 0; n < maxChars; n++ )
		if( chars[n] )
			numChars++;

	if( fontDescFormat == 0 )
		fprintf(f, "chars count=%d\r\n", numChars);
	else if( fontDescFormat == 1 )
		fprintf(f, "  <chars count=\"%d\">\r\n", numChars);
	else if( fontDescFormat == 2 )
	{
		fputc(4, f);

		// Determine the size of this block
		int size = 4+(2+2+2+2+2+2+2+2+1+1)*numChars;
		fwrite(&size, 4, 1, f);
	}

	for( n = 0; n < maxChars; n++ )
	{
        if( chars[n] )
		{
			int page, chnl;
			page = chars[n]->page;
			chnl = chars[n]->chnl;
			
			if( fontDescFormat == 1 )
				fprintf(f, "    <char id=\"%d\" x=\"%d\" y=\"%d\" width=\"%d\" height=\"%d\" xoffset=\"%d\" yoffset=\"%d\" xadvance=\"%d\" page=\"%d\" chnl=\"%d\" />\r\n", n, chars[n]->x, chars[n]->y, chars[n]->width, chars[n]->height, chars[n]->xoffset, chars[n]->yoffset, chars[n]->advance, page, chnl);
			else if( fontDescFormat == 0 )
				fprintf(f, "char id=%-4d x=%-5d y=%-5d width=%-5d height=%-5d xoffset=%-5d yoffset=%-5d xadvance=%-5d page=%-2d chnl=%-2d\r\n", n, chars[n]->x, chars[n]->y, chars[n]->width, chars[n]->height, chars[n]->xoffset, chars[n]->yoffset, chars[n]->advance, page, chnl);
			else
			{
#pragma pack(push)
#pragma pack(1)
				struct charBlock
				{
					WORD id;
					WORD x;
					WORD y;
					WORD width;
					WORD height;
					short xoffset;
					short yoffset;
					short xadvance;
					char  page;
					char  channel;
				} charInfo;
#pragma pack(pop)

				charInfo.id = n;
				charInfo.x  = chars[n]->x;
				charInfo.y  = chars[n]->y;
				charInfo.width = chars[n]->width;
				charInfo.height = chars[n]->height;
				charInfo.xoffset = chars[n]->xoffset;
				charInfo.yoffset = chars[n]->yoffset;
				charInfo.xadvance = chars[n]->advance;
				charInfo.page = page;
				charInfo.channel = chnl;

				fwrite(&charInfo, sizeof(charInfo), 1, f);
			}
		}
	}

	if( fontDescFormat == 1 )
		fprintf(f, "  </chars>\r\n");


	// Save the kerning pairs as well
	KERNINGPAIR *pairs = 0;
	int num = 0;
	if( useUnicode )
	{
		num = GetKerningPairsW(dc, 0, 0);
		pairs = new KERNINGPAIR[num];
		GetKerningPairsW(dc, num, pairs);
	}
	else
	{
		num = GetKerningPairs(dc, 0, 0);
		pairs = new KERNINGPAIR[num];
		GetKerningPairs(dc, num, pairs);
	}

	if( num > 0 )
	{
		// Count the number of kerning pairs
		int count = 0;
		for( int n = 0; n < num; n++ )
		{
			if( pairs[n].wFirst < maxChars && pairs[n].wSecond < maxChars &&
				!disabled[pairs[n].wFirst] && !disabled[pairs[n].wSecond] &&
				selected[pairs[n].wFirst] && selected[pairs[n].wSecond] &&
				pairs[n].iKernAmount/aa )
			{
				count++;
			}
		}

		if( fontDescFormat == 0 )
			fprintf(f, "kernings count=%d\r\n", count);
		else if( fontDescFormat == 1 )
			fprintf(f, "  <kernings count=\"%d\">\r\n", count);
		else if( fontDescFormat == 2 )
		{
			fputc(5, f);

			// Determine the size of the block
			int size = 4+count*6;
			fwrite(&size, 4, 1, f);
		}
	}

	// It's been reported that for Chinese WinXP the kerning pairs for 
	// non-unicode charsets may contain characters > 255, so we need to 
	// filter for this.
	for( n = 0; n < num; n++ )
	{
		if( pairs[n].wFirst < maxChars && pairs[n].wSecond < maxChars &&
			!disabled[pairs[n].wFirst] && !disabled[pairs[n].wSecond] &&
			selected[pairs[n].wFirst] && selected[pairs[n].wSecond] &&
			pairs[n].iKernAmount/aa )
		{
			if( fontDescFormat == 1 )
				fprintf(f, "    <kerning first=\"%d\" second=\"%d\" amount=\"%d\" />\r\n", pairs[n].wFirst, pairs[n].wSecond, pairs[n].iKernAmount/aa);
			else if( fontDescFormat == 0 )
				fprintf(f, "kerning first=%-3d second=%-3d amount=%-4d\r\n", pairs[n].wFirst, pairs[n].wSecond, pairs[n].iKernAmount/aa);
			else 
			{
#pragma pack(push)
#pragma pack(1)
				struct kerningBlock
				{
					WORD first;
					WORD second;
					short amount;
				} kerning;
#pragma pack(pop)
				kerning.first = pairs[n].wFirst;
				kerning.second = pairs[n].wSecond;
				kerning.amount = pairs[n].iKernAmount/aa;

				fwrite(&kerning, sizeof(kerning), 1, f);
			}
		}
	}

	if( pairs )
		delete[] pairs;

	if( num > 0 && fontDescFormat == 1 )
		fprintf(f, "  </kernings>\r\n");

	if( fontDescFormat == 1 ) fprintf(f, "</font>\r\n");

	fclose(f);

	SelectObject(dc, oldFont);
	DeleteObject(font);

	DeleteDC(dc);


	// Save the image file
	for( n = 0; n < (signed)pages.size(); n++ )
	{
		string str = acStringFormat("%s_%02d.%s", filename.c_str(), n, textureFormat.c_str());

		acImage::Image image;
		image.width = outWidth;
		image.height = outHeight;
		if( outBitDepth == 32 )
		{
			image.pitch = image.width*4;
			image.format = acImage::PF_A8R8G8B8;
		}
		else
		{
			image.pitch = image.width;
			image.format = acImage::PF_A8;
		}

		image.data = new BYTE[image.pitch * image.height];

		// Generate the output texture for saving
		pages[n]->GenerateOutputTexture();

		cImage *page = pages[n]->GetPageImage();
		if( outBitDepth == 8 )
		{
			// Write image data
			for( int y = 0; y < outHeight; y++ )
			{
				for( int x = 0; x < outWidth; x++ )
				{
					DWORD pixel = page->pixels[y*outWidth + x];
					image.data[y*image.pitch + x] = (BYTE)(pixel>>24);
				}
			}
		}
		else
		{
			// Write image data
			for( int y = 0; y < outHeight; y++ )
			{
				for( int x = 0; x < outWidth; x++ )
				{
					DWORD pixel = page->pixels[y*outWidth + x];
					*(DWORD*)&image.data[y*image.pitch + x*4] = pixel;
				}
			}
		}

		if( textureFormat == "tga" )
			acImage::SaveTga(str.c_str(), image);
		else if( textureFormat == "png" )
			acImage::SavePng(str.c_str(), image);
		else if( textureFormat == "dds" )
			acImage::SaveDds(str.c_str(), image);
	}

	return 0;
}

int CFontGen::SaveConfiguration(const char *szFile)
{
	string filename = szFile;
	if( stricmp(filename.substr(filename.length() - 5).c_str(), ".bmfc") == 0 )
		filename = filename.substr(0, filename.length() - 5);

	FILE *f = fopen((filename + ".bmfc").c_str(), "wb");
	if( f == 0 ) return -1;
	
	fprintf(f, "# AngelCode Bitmap Font Generator configuration file\n");
	fprintf(f, "fileVersion=%d\n", 1);

	fprintf(f, "\n# font settings\n");
	fprintf(f, "fontName=%s\n", fontName.c_str());
	fprintf(f, "charSet=%d\n", charSet);
	fprintf(f, "fontSize=%d\n", fontSize);
	fprintf(f, "aa=%d\n", aa);
	fprintf(f, "scaleH=%d\n", scaleH);
	fprintf(f, "useSmoothing=%d\n", useSmoothing);
	fprintf(f, "isBold=%d\n", isBold);
	fprintf(f, "isItalic=%d\n", isItalic);
	fprintf(f, "useUnicode=%d\n", useUnicode);
	fprintf(f, "disableBoxChars=%d\n", disableBoxChars);

	fprintf(f, "\n# character alignment\n");
	fprintf(f, "paddingDown=%d\n", paddingDown);
	fprintf(f, "paddingUp=%d\n", paddingUp);
	fprintf(f, "paddingRight=%d\n", paddingRight);
	fprintf(f, "paddingLeft=%d\n", paddingLeft);	
	fprintf(f, "spacingHoriz=%d\n", spacingHoriz);
	fprintf(f, "spacingVert=%d\n", spacingVert);

	fprintf(f, "\n# output file\n");
	fprintf(f, "outWidth=%d\n", outWidth);
	fprintf(f, "outHeight=%d\n", outHeight);
	fprintf(f, "outBitDepth=%d\n", outBitDepth);
	fprintf(f, "fontDescFormat=%d\n", fontDescFormat);
	fprintf(f, "fourChnlPacked=%d\n", fourChnlPacked);
	fprintf(f, "textureFormat=%s\n", textureFormat.c_str());
	fprintf(f, "textureCompression=%d\n", textureCompression);

	fprintf(f, "\n# outline\n");
	fprintf(f, "outlineThickness=%d\n", outlineThickness);
	fprintf(f, "is1ChnlEncoded=%d\n", is1ChnlEncoded);

	fprintf(f, "\n# selected chars\n");
	
	int maxChars = useUnicode ? 65536 : 256;
	int lastChar = -1;
	bool isRange = false;
	int lineLength = 0;
	for( int n = 0; n < maxChars; n++ )
	{
		if( selected[n] && !disabled[n] )
		{
			// Is this the first char on the line?
			if( lastChar == -1 )
				lineLength += fprintf(f, "chars=%d", n);
			// Is this a new start?
			else if( lastChar != n-1 )
				lineLength += fprintf(f, ",%d", n);
			// Is this a range?
			if( lastChar == n-1 )
			{
				if( !isRange ) 
				{
					lineLength += fprintf(f, "-");
					isRange = true;
				}
				// Is this the last char in the range?
				if( n+1 == maxChars || !selected[n+1] || disabled[n+1] )
				{
					lineLength += fprintf(f, "%d", n);
					isRange = false;
				}
			}

			lastChar = n;

			// Make sure the line doesn't get too long
			if( lineLength > 100 && !isRange )
			{
				fprintf(f, "\n");
				lineLength = 0;
				lastChar = -1;
			}
		}
	}
	fprintf(f, "\n");

	fprintf(f, "\n# imported icon images\n");
	for( n = 0; n < (signed)iconImages.size(); n++ )
	{
		fprintf(f, "icon=\"%s\",%d\n", iconImages[n]->fileName.c_str(), iconImages[n]->id);
	}

	fclose(f);

	return 0;
}

int CFontGen::LoadConfiguration(const char *filename)
{
	ClearIconImages();

	// Read the values into temporary variables 
	int    _fileVersion        = 0;
	string _fontName           = "Arial";
	int    _charSet            = ANSI_CHARSET;
	int    _fontSize           = 32;
	int    _aa                 = 1;
	int    _scaleH             = 100;
	bool   _useSmoothing       = true;
	bool   _isBold             = false;
	bool   _isItalic           = false;
	bool   _useUnicode         = true;
	int    _paddingDown        = 0;
	int    _paddingUp          = 0;
	int    _paddingRight       = 0;
	int    _paddingLeft        = 0;
	int    _spacingHoriz       = 1;
	int    _spacingVert        = 1;
	int    _outWidth           = 256;
	int    _outHeight          = 256;
	int    _outBitDepth        = 8;
	int    _fontDescFormat     = 0;
	bool   _fourChnlPacked     = false;
	string _textureFormat      = "tga";
	int    _textureCompression = 0;
    bool   _disableBoxChars    = true;
	int    _outlineThickness   = 0;
	bool   _is1ChnlEncoded     = false;
	
	bool   _selected[65536];
	memset(_selected, 0, sizeof(_selected));

	// Open the file
	FILE *f = fopen(filename, "rb");
	if( f == 0 ) return -1;

	// Read each line
	char line[1024];
	while( fgets(line, 1024, f) )
	{
		// Does the line contain a comment?
		char *c = strchr(line, '#');
		if( c ) *c = '\0';

		// Does the line contain a name-value pair?
		c = strchr(line, '=');
		if( c )
		{
			*c++ = '\0';
			if(      strcmp(line, "fileVersion")        == 0 ) _fileVersion        = atoi(c);
			else if( strcmp(line, "fontName")           == 0 ) _fontName           = c;
			else if( strcmp(line, "charSet")            == 0 ) _charSet            = atoi(c);
			else if( strcmp(line, "fontSize")           == 0 ) _fontSize           = atoi(c);
			else if( strcmp(line, "aa")                 == 0 ) _aa                 = atoi(c);
			else if( strcmp(line, "scaleH")             == 0 ) _scaleH             = atoi(c);
			else if( strcmp(line, "useSmoothing")       == 0 ) _useSmoothing       = atoi(c) ? true : false;
			else if( strcmp(line, "isBold")             == 0 ) _isBold             = atoi(c) ? true : false;
			else if( strcmp(line, "isItalic")           == 0 ) _isItalic           = atoi(c) ? true : false;
			else if( strcmp(line, "useUnicode")         == 0 ) _useUnicode         = atoi(c) ? true : false;
			else if( strcmp(line, "paddingDown")        == 0 ) _paddingDown        = atoi(c);
			else if( strcmp(line, "paddingUp")          == 0 ) _paddingUp          = atoi(c);
			else if( strcmp(line, "paddingRight")       == 0 ) _paddingRight       = atoi(c);
			else if( strcmp(line, "paddingLeft")        == 0 ) _paddingLeft        = atoi(c);
			else if( strcmp(line, "spacingHoriz")       == 0 ) _spacingHoriz       = atoi(c);
			else if( strcmp(line, "spacingVert")        == 0 ) _spacingVert        = atoi(c);
			else if( strcmp(line, "outWidth")           == 0 ) _outWidth           = atoi(c);
			else if( strcmp(line, "outHeight")          == 0 ) _outHeight          = atoi(c);
			else if( strcmp(line, "outBitDepth")        == 0 ) _outBitDepth        = atoi(c);
			else if( strcmp(line, "fontDescFormat")     == 0 ) _fontDescFormat     = atoi(c);
			else if( strcmp(line, "fourChnlPacked")     == 0 ) _fourChnlPacked     = atoi(c) ? true : false;
			else if( strcmp(line, "textureFormat")      == 0 ) _textureFormat      = c;
			else if( strcmp(line, "textureCompression") == 0 ) _textureCompression = atoi(c);
			else if( strcmp(line, "disableBoxChars")    == 0 ) _disableBoxChars    = atoi(c) ? true : false;
			else if( strcmp(line, "outlineThickness")   == 0 ) _outlineThickness   = atoi(c);
			else if( strcmp(line, "is1ChnlEncoded")     == 0 ) _is1ChnlEncoded     = atoi(c) ? true : false;
			else if( strcmp(line, "chars") == 0 )
			{
				// Parse the ranges
				while( *c >= '0' && *c <= '9' )
				{
					int firstChar = strtol(c, &c, 10);
					if( *c == '-' )
					{
						int lastChar = strtol(c+1, &c, 10);
						for( int n = firstChar; n <= lastChar; n++ )
							_selected[n] = true;
					}
					else
						_selected[firstChar] = true;

					if( *c == ',' ) c++;
				}
			}
			else if( strcmp(line, "icon") == 0 )
			{
				// Parse the icon image
				char *start = strchr(c, '"');
				char *end = 0;
				if( start )
					end = strchr(start+1, '"');
				if( end )
				{
					string file;
					file.assign(start+1, end);
					
					int id = 0;
					if( *(end+1) == ',' )
					{
						c = end+2;
						id = strtol(c, &c, 10);
					}

					AddIconImage(file.c_str(), id);
				}
			}
		}
	}

	fclose(f);

	// Make sure the values are in valid ranges
	size_t pos = _fontName.find_last_not_of(" \t\n\r");
	if( pos != string::npos ) _fontName.erase(pos + 1);
	if( _aa < 1 ) _aa = 1; if( _aa > 4 ) _aa = 4;
	if( _scaleH < 1 ) _scaleH = 1;
	if( _paddingDown < 0 ) _paddingDown = 0;
	if( _paddingUp < 0 ) _paddingUp = 0;
	if( _paddingRight < 0 ) _paddingRight = 0;
	if( _paddingLeft < 0 ) _paddingLeft = 0;
	if( _spacingHoriz < 0 ) _spacingHoriz = 0;
	if( _spacingVert < 0 ) _spacingVert = 0;
	if( _outWidth < 1 ) _outWidth = 1;
	if( _outHeight < 1 ) _outHeight = 1;
	if( _outBitDepth != 8 && _outBitDepth != 32 ) _outBitDepth = 8;
	if( _fontDescFormat < 0 || _fontDescFormat > 2 ) _fontDescFormat = 0;
    
	pos = _textureFormat.find_last_not_of(" \t\n\r");
	if( pos != string::npos ) _textureFormat.erase(pos + 1);

	if( _textureFormat == "tga" )
	{
		_textureCompression = 0;
	}
	else if( _textureFormat == "png" )
	{
		_textureCompression = 0;
	}
	else if( _textureFormat == "dds" )
	{
		_textureCompression = 0;
//		if( _textureCompression < 0 ) _textureCompression = 0;
//		if( _textureCompression > 3 ) _textureCompression = 3;
	}
	else
	{
		_textureFormat      = "tga";
		_textureCompression = 0;
	}

	// Is it the right file version?
	if( _fileVersion != 1 )
		return -1;

	// Set the properties
	SetFontName(_fontName);
	SetCharSet(_charSet);
	SetFontSize(_fontSize);
	SetAntiAliasingLevel(_aa);
	SetScaleHeight(_scaleH);
	SetUseSmoothing(_useSmoothing);
	SetBold(_isBold);
	SetItalic(_isItalic);
	SetUseUnicode(_useUnicode);
	SetPaddingDown(_paddingDown);
	SetPaddingUp(_paddingUp);
	SetPaddingRight(_paddingRight);
	SetPaddingLeft(_paddingLeft);
	SetSpacingHoriz(_spacingHoriz);
	SetSpacingVert(_spacingVert);
	SetOutWidth(_outWidth);
	SetOutHeight(_outHeight);
	SetOutBitDepth(_outBitDepth);
	SetFontDescFormat(_fontDescFormat);
	Set4ChnlPacked(_fourChnlPacked);
	SetBoxCharsDisabled(_disableBoxChars);
	SetTextureFormat(_textureFormat);
	SetTextureCompression(_textureCompression);
	SetOutlineThickness(_outlineThickness);
	Set1ChnlEncoded(_is1ChnlEncoded);

	Prepare();

	int maxChars = useUnicode ? 65536 : 256;
	for( int n = 0; n < maxChars; n++ )
		SetSelected(n, _selected[n]);

	return 0;
}

int CFontGen::GetNumFailedChars()
{
	int numFailed = 0;
	for( int n = 0; n < 65536; n++ )
		if( noFit[n] ) numFailed++;

	return numFailed;
}

int CFontGen::SelectCharsFromFile(const char *filename)
{
	if( isWorking ) return -1;

	FILE *f = fopen(filename, "rb");
	if( f == 0 ) return -1;

	memset(noFit, 0, sizeof(noFit));

	if( IsUsingUnicode() )
	{
		wchar_t buf[1024];
		size_t cnt;
		while( (cnt = fread(buf, 2, 1024, f)) )
			for( UINT n = 0; n < cnt; n++ )
			{
				if( SetSelected(buf[n], true) < 0 )
				{
					// Control characters are not visible
					if( buf[n] >= 32 )
						noFit[buf[n]] = true;
				}
			}
	}
	else
	{
		char buf[1024];
		size_t cnt;
		while( (cnt = fread(buf, 1, 1024, f)) )
			for( UINT n = 0; n < cnt; n++ )
			{
				if( SetSelected(buf[n], true) < 0 )
					noFit[buf[n]] = true;
			}
	}

	fclose(f);

	return 0;
}

int CFontGen::FindNextFailedCharacterSubset(int startSubset)
{
	if( useUnicode )
	{
		// What is the start of the next subset?
		startSubset++;
		if( startSubset >= numUnicodeSubsets )
			startSubset = 0;

		int startChar = UnicodeSubsets[startSubset].range[0];
		for( int n = startChar; n < 65536; n++ )
		{
			if( noFit[n] )
				return SubsetFromChar(n);
		}

		for( n = 0; n <= startChar; n++ )
		{
			if( noFit[n] )
				return SubsetFromChar(n);
		}
	}

	return 0;
}

int CFontGen::SubsetFromChar(int ch)
{
	if( useUnicode )
	{
		for( int n = 0; n < numUnicodeSubsets; n++ )
		{
			if( ch >= UnicodeSubsets[n].range[0] &&
				ch <= UnicodeSubsets[n].range[1] )
				return n;
		}
	}

	return 0;
}

void CFontGen::ClearFailedCharacters()
{
	memset(noFit, 0, sizeof(noFit));
}