#include <algorithm>

#include "fontpage.h"
#include "fontchar.h"
#include "fontgen.h"

#define CLR_BORDER 0x007F00ul
#define CLR_UNUSED 0xFF0000ul

CFontPage::CFontPage(CFontGen *gen, int id, int width, int height, int spacingH, int spacingV)
{
	this->gen = gen;
	pageId = id;

	// Allocate and clear the image with the unused color
	pageImg = new cImage(width, height);
	pageImg->Clear(CLR_UNUSED);

	// Initialize the height array that shows free space
	heights[0] = new int[width];
	memset(heights[0], 0, width*4);
	heights[1] = 0;
	heights[2] = 0;
	heights[3] = 0;

	currX = 0;

	this->spacingH = spacingH;
	this->spacingV = spacingV;

	paddingRight  = 0;
	paddingLeft   = 0;
	paddingUp     = 0;
	paddingDown   = 0;
}

CFontPage::~CFontPage()
{
	if( pageImg )
		delete pageImg;

	for( int n = 0; n < 4; n++ )
		if( heights[n] )
			delete[] heights[n];
}

void CFontPage::SetIntendedFormat(int bitDepth, bool fourChnlPacked, bool is1ChnlEncoded)
{
	this->bitDepth = bitDepth;
	this->fourChnlPacked = fourChnlPacked;
	this->is1ChnlEncoded = is1ChnlEncoded;
}

void CFontPage::AddChar(int cx, int cy, CFontChar *ch, int channel)
{
	// Update the charInfo with the extra draw rect
	ch->x = cx;
	ch->y = cy;
	ch->page = pageId;
	ch->width += paddingLeft + paddingRight;
	ch->height += paddingUp + paddingDown;
	ch->xoffset -= paddingLeft;
	ch->yoffset -= paddingUp;
	if( bitDepth == 32 && fourChnlPacked )
        ch->chnl = !ch->FitsInOneChannel(is1ChnlEncoded) ? 0xF : (1<<channel);
	else
		ch->chnl = 0xF;

	// Update heights
	cImage *img = ch->charImg;
	for( int x = -spacingH; x < img->width + paddingLeft + paddingRight + spacingH; x++ )
	{
		int tempX = x + cx;
		if( tempX < 0 ) tempX += pageImg->width;
		if( cy + img->height + spacingV + paddingUp + paddingDown > heights[channel][tempX] )
			heights[channel][tempX] = cy + img->height + spacingV + paddingUp + paddingDown;
	}

	chars.push_back(ch);

	// Increment counter in CFontGen
	gen->counter++;
}

int CFontPage::AddChar(CFontChar *ch, int channel)
{
	int origX = currX;
	cImage *img = ch->charImg;

	// Iterate for each possible x position
	int i = 0;
	while( i++ < pageImg->width - img->width - paddingRight - paddingLeft - spacingH )
	{
		// Is the character narrow enough to fit?
		if( img->width + currX + paddingRight + paddingLeft > pageImg->width - spacingH )
		{
			// Start from the left side again
			currX = 0;
		}

		// Will the character fit in this place?
		int cy = 0;
		for( int n = 0; n < img->width + paddingLeft + paddingRight; n++ )
		{
			if( heights[channel][n+currX] > cy ) 
				cy = heights[channel][n+currX];
		}

		if( cy + img->height + paddingUp + paddingDown <= pageImg->height - spacingV )
		{
			// Are we creating any holes?
			for( int x = 0; x < img->width + paddingLeft + paddingRight; x++ )
			{
				int tempX = x + currX;
				if( cy - spacingV > heights[channel][tempX] )
				{
					SHole hole;
					hole.x    = tempX;
					hole.y    = heights[channel][tempX];
					hole.w    = img->width + paddingLeft + paddingRight - x;
					hole.h    = cy - spacingV - hole.y;
					hole.chnl = channel;
					holes.push_back(hole);
					break;
				}
			}

			AddChar(currX, cy, ch, channel);

			currX += img->width + spacingH + paddingLeft + paddingRight;

			return 0;
		}
		else
		{
			currX++;
		}
	}

	currX = origX;
	return -1;
}

cImage *CFontPage::GetPageImage()
{
	return pageImg;
}

void CFontPage::SetPadding(int left, int up, int right, int down)
{
	paddingLeft  = left;
	paddingUp    = up;
	paddingRight = right;
	paddingDown  = down;
}

int CFontPage::GetNextIdealImageWidth()
{
	return pageImg->width - currX - paddingRight - paddingLeft - spacingH;
}

void CFontPage::GeneratePreviewTexture(int channel)
{
	pageImg->Clear(CLR_UNUSED);

	// Copy the font char images to the texture
	for( unsigned int n = 0; n < chars.size(); n++ )
	{
		if( chars[n]->chnl & (1<<channel) )
		{
			int cx = chars[n]->x + paddingLeft;
			int cy = chars[n]->y + paddingUp;
			cImage *img = chars[n]->charImg;
	
			if( chars[n]->HasOutline() )
			{
				// Show the outline, by blending against blue background
				for( int y = 0; y < img->height; y++ )
				{
					for( int x = 0; x < img->width; x++ )
					{
						DWORD p = img->pixels[y*img->width+x];
						if( (p >> 24) < 0xFF )
							p += 255 - (p>>24);
						pageImg->pixels[(y+cy)*pageImg->width+(x+cx)] = p;
					}
				}
			}
			else
			{
				for( int y = 0; y < img->height; y++ )
				{
					for( int x = 0; x < img->width; x++ )
						pageImg->pixels[(y+cy)*pageImg->width+(x+cx)] = img->pixels[y*img->width+x];
				}
			}

			// Draw the spacing borders
			if( spacingH > 0 )
			{
				int cx1 = chars[n]->x - 1;
				if( cx1 < 0 ) cx1 += pageImg->width;
				int cx2 = chars[n]->x + chars[n]->width;
				if( cx2 >= pageImg->width ) cx2 -= pageImg->width;
				int cy = chars[n]->y;

				for( int y = 0; y < chars[n]->height; y++ )
				{
					pageImg->pixels[(cy+y)*pageImg->width+cx1] = CLR_BORDER;
					pageImg->pixels[(cy+y)*pageImg->width+cx2] = CLR_BORDER;
				}
			}

			if( spacingV > 0 )
			{
				int cy1 = chars[n]->y - 1;
				if( cy1 < 0 ) cy1 += pageImg->height;
				int cy2 = chars[n]->y + chars[n]->height;
				if( cy2 >= pageImg->height ) cy2 -= pageImg->height;
				int cx = chars[n]->x;

				for( int x = 0; x < chars[n]->width; x++ )
				{
					pageImg->pixels[cy1*pageImg->width+x+cx] = CLR_BORDER;
					pageImg->pixels[cy2*pageImg->width+x+cx] = CLR_BORDER;
				}
			}
		}
	}
}

void CFontPage::GenerateOutputTexture()
{
	// Clear the image
	memset(pageImg->pixels, 0, pageImg->width*pageImg->height*4);

	// Copy the font char images to the texture
	for( unsigned int n = 0; n < chars.size(); n++ )
	{
		int cx = chars[n]->x + paddingLeft;
		int cy = chars[n]->y + paddingUp;
		cImage *img = chars[n]->charImg;

		if( !chars[n]->FitsInOneChannel(is1ChnlEncoded) )
		{
			for( int y = 0; y < img->height; y++ )
			{
				for( int x = 0; x < img->width; x++ )
					pageImg->pixels[(y+cy)*pageImg->width+(x+cx)] = img->pixels[y*img->width+x];
			}
		}
		else
		{
			if( fourChnlPacked )
			{
				for( int y = 0; y < img->height; y++ )
				{
					for( int x = 0; x < img->width; x++ )
					{
						DWORD p = chars[n]->GetPixelValue(x, y);
						if( chars[n]->chnl == 2 )
							p <<= 8;
						else if( chars[n]->chnl == 4 )
							p <<= 16;
						else if( chars[n]->chnl == 8 )
							p <<= 24;
							
						pageImg->pixels[(y+cy)*pageImg->width+(x+cx)] |= p;
					}
				}
			}
			else
			{
				for( int y = 0; y < img->height; y++ )
				{
					for( int x = 0; x < img->width; x++ )
					{
						DWORD p = ((BYTE)chars[n]->GetPixelValue(x, y)<<24)|0xFFFFFF;
						pageImg->pixels[(y+cy)*pageImg->width+(x+cx)] = p;
					}
				}
			}
		}
	}
}

// This global pointer will be used by the sorting algorithm
CFontChar **g_chars = 0;

// Define a structure and comparison operator for the sorting algorithm
struct element { int index; };
bool operator<(const element &a, const element &b) 
{
	// We want to sort the characters from larger to smaller
	if( g_chars[a.index]->height > g_chars[b.index]->height ||
	    (g_chars[a.index]->height == g_chars[b.index]->height &&
	     g_chars[a.index]->width > g_chars[b.index]->width) )
		return true;
	return false;
}

void CFontPage::SortList(CFontChar **chars, int *index, int numChars)
{
	g_chars = chars;
	std::sort((element*)index, (element*)(&index[numChars]));
}

void CFontPage::AddChars(CFontChar **chars, int maxChars)
{
	// Add the colored images first
	AddCharsToPage(chars, maxChars, true, 0);

	// Check if we should stop
	if( gen->stopWorking ) return;

	// Duplicate the height array for the other channels
	for( int n = 1; n < 4; n++ )
	{
		heights[n] = new int[pageImg->width];
		memcpy(heights[n], heights[0], pageImg->width*sizeof(int));
	}

	// Remove the current holes
	holes.resize(0);

	// Then the black & white images
	currX = 0;
	AddCharsToPage(chars, maxChars, false, 0);

	// Check if we should stop
	if( gen->stopWorking ) return;

	if( fourChnlPacked )
	{
		for( int n = 1; n < 4; n++ )
		{
			currX = 0;
			AddCharsToPage(chars, maxChars, false, n);

			// Check if we should stop
			if( gen->stopWorking ) return;
		}
	}
}

void CFontPage::AddCharsToPage(CFontChar **chars, int maxChars, bool colored, int channel)
{
	static int indexA[65536], indexB[65536];
	int *index = indexA, *index2 = indexB;
	int numChars = 0, numChars2 = 0;

	// Add images to the list
	for( int n = 0; n < maxChars; n++ )
	{
		if( chars[n] && !chars[n]->FitsInOneChannel(is1ChnlEncoded) == colored )
			index[numChars++] = n;
	}

	// Sort the characters by height/width, largest first
	SortList(chars, index, numChars);

	// Add the images to the page
	while( numChars > 0 )
	{
		// Fill holes
		for( int h = 0; h < (signed)holes.size(); h++ )
		{
			int bestMatch;
			bestMatch = -1;

			// Find the best matching character to fill the hole
			for( int n = 0; n < numChars; n++ )
			{
				if( (holes[h].w == chars[index[n]]->charImg->width + paddingLeft + paddingRight) &&
					(holes[h].h == chars[index[n]]->charImg->height + paddingUp + paddingDown) )
				{
					bestMatch = n;
					break;
				}
				else if( (holes[h].w >= chars[index[n]]->charImg->width + paddingLeft + paddingRight) &&
					     (holes[h].h >= chars[index[n]]->charImg->height + paddingUp + paddingDown) )
				{
					if( bestMatch != -1 )
					{
						if( (chars[index[n]]->charImg->width > chars[index[bestMatch]]->charImg->width) ||
							(chars[index[n]]->charImg->height > chars[index[bestMatch]]->charImg->height) )
							bestMatch = n;
					}
					else
						bestMatch = n;
				}
			}

			bool remove = true;
			if( bestMatch != -1 )
			{
				int x = holes[h].x;
				int y = holes[h].y;

				// There may still be room for more 
				if( holes[h].w - spacingH > chars[index[bestMatch]]->charImg->width + paddingLeft + paddingRight )
				{
					holes[h].x += chars[index[bestMatch]]->charImg->width + paddingLeft + paddingRight + spacingH;
					holes[h].w -= chars[index[bestMatch]]->charImg->width + paddingLeft + paddingRight + spacingH;
					remove = false;
				}
				else if( holes[h].h - spacingV > chars[index[bestMatch]]->charImg->height + paddingUp + paddingDown )
				{
					holes[h].y += chars[index[bestMatch]]->charImg->height + paddingUp + paddingDown + spacingV;
					holes[h].h -= chars[index[bestMatch]]->charImg->height + paddingUp + paddingDown + spacingV;
					remove = false;
				}

				AddChar(x, y, chars[index[bestMatch]], channel);
				chars[index[bestMatch]] = 0;

				// Check if we should stop
				if( gen->stopWorking ) return;

				// Compact the list 
				numChars--;
				for( int n = bestMatch; n < numChars; n++ )
					index[n] = index[n+1];
			}
			
			// Remove the hole
			if( remove )
			{
				if( h < (signed)holes.size() - 1 )
					holes[h] = holes[holes.size()-1];
				holes.pop_back();
			}
			h--;
		}

		numChars2 = 0;
		bool allTooWide = true;
		bool drawn = false;

		// Add chars to texture
		for( int n = 0; n < numChars; n++ )
		{
			bool ok = false;
			if( chars[index[n]]->charImg->width <= GetNextIdealImageWidth() )
			{
				allTooWide = false;
				int r = AddChar(chars[index[n]], channel);
				if( r >= 0 )
				{
					chars[index[n]] = 0;
					ok = true;
					drawn = true;

					// Check if we should stop
					if( gen->stopWorking ) return;
				}
			}

			if( !ok )
			{
				// Move it to the next index list
				index2[numChars2++] = index[n];
			}
		}

		if( index == indexA )
		{
			// Swap indices
			index = indexB;
			index2 = indexA;
		}
		else
		{
			index = indexA;
			index2 = indexB;
		}

		numChars = numChars2;
		numChars2 = 0;

		if( allTooWide )
		{
			// Start over from the left side
			currX = 0;
		}
		else if( !drawn )
		{
			// Next page
			break;
		}
	}
}